@@ -1,15 +1,18 @@
+from collections import namedtuple
+from functools import partial
 from math import floor
+from multiprocess import Pool   # Uses dill for serialising.
 from multiprocessing import shared_memory
 import os
 import re
 import shutil
-from typing import Union
+from typing import List, Mapping, Tuple, Union
 
 import numpy as np
 import pandas as pd
 
 from expam.classify import ResultsPathConfig
-from expam.classify.config import load_results_config
+from expam.classify.config import create_tax_results, load_results_config
 from expam.classify.process import ExpamClassifierProcesses
 from expam.classify.taxonomy import TaxonomyNCBI
 from expam.database import EXPAM_TEMP_EXT, TIMEOUT, UNION_RATIO, FileLocationConfig
@@ -29,14 +32,15 @@ def run_classifier(
     k: int,
     n: int,
     phylogeny_path: str,
-    keys_shape: tuple[int],
-    values_shape: tuple[int],
+    keys_shape: Tuple[int],
+    values_shape: Tuple[int],
     logging_dir: str,
-    taxonomy: bool = False, cutoff: float = 0.0, groups: Union[None, list[tuple]] = None,
+    taxonomy: bool = False, cutoff: float = 0.0, groups: Union[None, List[tuple]] = None,
     keep_zeros: bool = False, cpm: float = 0.0, use_node_names: bool = True,
-    phyla: bool = False, name_taxa: Union[dict[str, str], None] = None,
-    colour_list: list[str] = None, paired_end: bool = False, alpha: float = 1.0,
-    log_scores: bool = False, itol_mode: bool = False
+    phyla: bool = False, name_taxa: Union[Mapping[str, str], None] = None,
+    colour_list: List[str] = None, paired_end: bool = False, alpha: float = 1.0,
+    log_scores: bool = False, itol_mode: bool = False,
+    sep: str = ","
 ):
     output_config: ResultsPathConfig = load_results_config(out_dir, create=True)
     database_config: FileLocationConfig = load_database_config(db_dir)
@@ -63,11 +67,13 @@ def run_classifier(
             logging_dir=logging_dir,
             temp_dir=output_config.temp,
             paired_end=paired_end,
-            alpha=alpha
+            alpha=alpha,
+            sep=sep,
+            n_processes=n
         )
 
         # Run the classification with n processes.
-        d.run(n)
+        d.run()
 
         # Load accession ids.
         sequence_ids = list(yield_csv(database_config.accession_id))
@@ -93,7 +99,8 @@ def run_classifier(
             phyla=phyla,
             name_taxa=name_taxa,
             colour_list=colour_list,
-            log_scores=log_scores
+            log_scores=log_scores,
+            sep=sep
         )
 
         if taxonomy:
@@ -137,9 +144,14 @@ def name_to_id(phylogeny_path: str):
     return index, index_map
 
 
+ResultsColumns = namedtuple("ResultsColumns", ["c_perc", "c_cumul", "c_count", "s_perc", "s_cumul", "s_count"])
+
 class Distribution:
+    _cols = ResultsColumns("Cumulative Classified Percentage", "Cumulative Classified Count", "Raw Classified Count",
+                           "Cumulative Split Percentage", "Cumulative Split Count", "Raw Split Count")
+
     def __init__(self, k, kmer_db, index: Index, lca_matrix, read_paths, out_dir, temp_dir, logging_dir, alpha,
-                 keep_zeros=False, cutoff=0.0, cpm=0.0, paired_end=False):
+                 keep_zeros=False, cutoff=0.0, cpm=0.0, paired_end=False, sep=",", n_processes: int = 1):
         """
 
         :param k: Int
@@ -165,10 +177,14 @@ class Distribution:
         self.cpm = cpm
 
         # Url containing reads.
+        self.results_config: ResultsPathConfig = load_results_config(out_dir, create=True)
         self.read_paths = read_paths
         self.paired_end = paired_end
 
-    def run(self, n):
+        self.SEP = sep
+        self.n_processes = n_processes
+
+    def run(self):
         """
         Begin processing reads in self.url.
         We have some collection of processes accessing reads from storage
@@ -176,10 +192,8 @@ class Distribution:
 
         """
         file_queue = self.prepare_queue(self.paired_end)
-        n_classifiers = floor(UNION_RATIO * n)
-        n_readers = n - n_classifiers
-
-        results_config: ResultsPathConfig = load_results_config(self.out_dir, create=True)
+        n_classifiers = floor(UNION_RATIO * self.n_processes)
+        n_readers = self.n_processes - n_classifiers
 
         mp_config = {
             "name": "expam",
@@ -198,7 +212,7 @@ class Distribution:
                         "k": self.k,
                         "kmer_db": self.kmer_db,
                         "lca_matrix": self.lca_matrix,
-                        "out_dir": results_config.phy,
+                        "out_dir": self.results_config.phy,
                         "temp_dir": self.temp_dir,
                         "n_readers": n_readers,
                         "logging_dir": self.logging_dir,
@@ -221,7 +235,11 @@ class Distribution:
 
         process_network.run()
 
-        # Combine raw read output.
+        # Process and combine outputs.
+        output_files = self.combine_raw_read_output()
+        self.create_sample_summaries(output_files)
+
+    def combine_raw_read_output(self):
         temporary_files = ls(self.temp_dir, ext=".reads_%s" % EXPAM_TEMP_EXT)
         base_file_names = {
             re.match(r'(\S+)_\d+.reads_%s' % EXPAM_TEMP_EXT, os.path.basename(file_name)).group(1)
@@ -229,145 +247,142 @@ class Distribution:
         }
 
         result_files = []
-        for base_file in base_file_names:
-            results_file_name = os.path.join(results_config.phy_raw, base_file + ".csv")
-            raw_files = [file for file in temporary_files if re.match(r'.*{base}_\d+.reads_{ext}$'.format(base=base_file, ext=EXPAM_TEMP_EXT), file)] 
-            results = [self.get_data(file) for file in raw_files]
+        for sample_name in base_file_names:
+            results_file_name = os.path.join(self.results_config.phy_raw, sample_name + ".csv")
+            raw_files = [file for file in temporary_files if re.match(r'.*{base}_\d+.reads_{ext}$'.format(base=sample_name, ext=EXPAM_TEMP_EXT), file)] 
+            
+            self.concatenate_files(results_file_name, *raw_files)
+            result_files.append((results_file_name, sample_name))
 
-            with open(results_file_name, "w") as f:
-                f.write("\n".join(results))
+        return result_files
 
-            result_files.append((results_file_name, base_file))
-
-        # Count reads at each node, for each sample.
-        counts_dicts = {
-            "C": {
-                "out_dir": results_config.phy_classified,
-                "counts": {}
-            },
-            "S": {
-                "out_dir": results_config.phy_split,
-                "counts": {}
-            }
-        }
-        counts_dicts["U"] = counts_dicts["C"]
-
-        for result_file, sample_name in result_files:
-            counts_dicts["C"]["counts"][sample_name] = [0 for _ in range(len(self.node_names))]
-            counts_dicts["S"]["counts"][sample_name] = [0 for _ in range(len(self.node_names))]
-
-            with open(result_file, "r") as f:
+    def create_sample_summaries(self, result_files):
+        nprocs = min(self.n_processes, len(result_files))
+        ###
+        ### Import raw counts into DataFrame.
+        def _raw_sample_to_series(results_file, sample_name, node_names):
+            sl = pd.Series(0, index=node_names, name=sample_name)
+            ml = pd.Series(0, index=node_names, name=sample_name)
+            with open(results_file, 'r') as f:
                 for line in f:
-                    data = line.split("\t")
-                    counts_dicts[data[0]]["counts"][sample_name][int(data[2].lstrip("p"))] += 1
+                    parts = line.split("\t")
+                    idx = int(parts[2].lstrip("p"))
+                    if parts[0] == "S":
+                        ml.iloc[idx] += 1
+                    else:
+                        sl.iloc[idx] += 1
+            return sample_name, sl, ml
 
-        # Get non-zero rows.
-        non_zero_indices = list({
-            i
-            for _, sample_name in result_files
-            for result_type in ("C", "S")
-            for i, count in enumerate(counts_dicts[result_type]["counts"][sample_name])
-            if (count > 0) or (i == 0)
-        })
+        ###
+        with Pool(processes=nprocs) as mp_pool:
+            raw_sample_series = sorted(
+                mp_pool.starmap_async(
+                    partial(_raw_sample_to_series, node_names=self.node_names),
+                    result_files).get(),
+                key = lambda t: t[0])   # Sorted in order of sample names.
+        
+        ###
+        ###
 
-        def get_cumulative(lst):
-            cumulative_counts = lst.copy()
+        ###
+        ### Create sample-wise summaries.
+        def remove_zeros(*dfs):
+            mask = None
+            for df in dfs:
+                non_zero = df.apply(lambda r: ((r > 0).any()) | (r.name == 'unclassified'), axis=1)
+                mask = non_zero if mask is None else mask | non_zero
+            for df in dfs:
+                yield df.loc[mask, :]
 
-            for i in range(len(self.node_names) - 1, 0, -1):
-                cumulative_counts[i] += sum([
-                    cumulative_counts[self.index._pointers[child]]
-                    for child in self.index.pool[i].children
-                ])
+        ###
+        def _sample_summary(
+            sample_name: str, sl: pd.Series, ml: pd.Series, out_base: str,
+            node_names: List[str], cols: ResultsColumns, index: Index,
+            cutoff: int, cpm: float, sep: str):
+            df = pd.DataFrame(0, index=node_names, columns=tuple(cols))
+            df.loc[:, cols.c_cumul] = sl.loc[:]
+            df.loc[:, cols.c_count] = sl.loc[:]
+            df.loc[:, cols.s_cumul] = ml.loc[:]
+            df.loc[:, cols.s_count] = ml.loc[:]
 
-            return cumulative_counts
-
-        # Sample-wise summary of counts in kraken-like format.
-        summary_format = "%f%%\t%d\t%d"
-
-        # Format phylogenetic node string.
-        self.node_names = [
-            "p" + self.node_names[i]
-            if self.index.pool[i].type == "Branch" and i > 0
-            else self.node_names[i]
-            for i in range(len(self.node_names))
-        ]
-
-        for _, sample_name in result_files:
-            total_reads = sum(counts_dicts["C"]["counts"][sample_name]) + sum(counts_dicts["S"]["counts"][sample_name])
-
-            cumulative_classifications = get_cumulative(counts_dicts["C"]["counts"][sample_name])
-            cumulative_splits = get_cumulative(counts_dicts["S"]["counts"][sample_name])
+            # Make counts cumulative.
+            cumulative_columns = [cols.c_cumul, cols.s_cumul]
+            for node in index.pool[1:][::-1]:
+                df.loc[node.name, cumulative_columns] += df.loc[node.children, cumulative_columns].sum(axis=0)
 
             # Employ cutoff.
-            if self.cutoff > 0 or self.cpm > 0:
-                cutoff = max((total_reads / 1e6) * self.cpm, self.cutoff)
-                valid_inds = list({
-                    i
-                    for cumulative_count in (cumulative_classifications, cumulative_splits)
-                    for i, count in enumerate(cumulative_count)
-                    if (count >= cutoff) or (i == 0)
-                })
-                valid_inds.sort()
-
-            else:
-                valid_inds = list(range(len(self.node_names)))
-
-            sample_summary = ""
-            for i in valid_inds:
-                if i not in non_zero_indices and not self.keep_zeros:
-                    continue
-
-                if i == 0:
-                    line_summary = "unclassified\t" + summary_format % (
-                        round(100 * cumulative_classifications[0] / total_reads, 3),
-                        cumulative_classifications[0],
-                        counts_dicts["C"]["counts"][sample_name][0]) + ("\t" * 3) + "\n"
-                else:
-                    line_classified_summary = self.node_names[i] + "\t" + summary_format % (
-                        round(100 * cumulative_classifications[i] / total_reads, 3),
-                        cumulative_classifications[i],
-                        counts_dicts["C"]["counts"][sample_name][i])
-                    line_splits_summary = summary_format % (
-                        round(100 * cumulative_splits[i] / total_reads, 3),
-                        cumulative_splits[i],
-                        counts_dicts["S"]["counts"][sample_name][i])
-                    line_summary = line_classified_summary + "\t" + line_splits_summary + "\n"
-
-                sample_summary += line_summary
-
-            with open(os.path.join(results_config.phy, sample_name + ".csv"), "w") as f:
-                f.write(sample_summary)
-
-        # Save raw counts to disk.
-        do_unclassified = True
-        for counts_dict in (counts_dicts["C"], counts_dicts["S"]):
-            out_dir, counts = counts_dict["out_dir"], counts_dict["counts"]
-            df = pd.DataFrame.from_dict(counts)
+            total_reads = df.loc[:, cols.c_count].sum() + df.loc[:, cols.s_count].sum()
+            if cutoff > 0 or cpm > 0.0:
+                cutoff = max((total_reads / 1e6) * cpm, cutoff)
+                mask = (df.loc[:, cols.c_cumul] < cutoff) & (df.loc[:, cols.s_cumul])
+                mask[0] = True  # Keep unclassified counts.
+                df.loc[mask, :] = 0
 
+            # Check for removal of zero rows.
             if not self.keep_zeros:
-                df = df.iloc[non_zero_indices]
-                index = [self.node_names[i] for i in non_zero_indices]
-            else:
-                index = self.node_names
+                (df, ) = remove_zeros(df)
 
-            if "unclassified" in index and not do_unclassified:
-                df.drop(0, inplace=True)
-                index = index[1:]
+            # Insert percentage form.
+            df.loc[:, cols.c_perc] = round(100 * df.loc[:, cols.c_cumul] / total_reads, 3).map(str) + "%"
+            df.loc[:, cols.s_perc] = round(100 * df.loc[:, cols.s_cumul] / total_reads, 3).map(str) + "%"
 
-            df.index = index
-            df.to_csv(out_dir, sep="\t")
+            # Write to disk.
+            out_file = os.path.join(out_base, sample_name + ".csv")
+            fixer = lambda name: 'p' + name if (name != "unclassified") and (index[name].type == "Branch") else name
+            df.index = [fixer(n) for n in df.index]
+            df.to_csv(out_file, sep=sep)
 
-            do_unclassified = False
+        ###
+        ###
+
+        with Pool(processes=nprocs) as mp_pool:
+            f = partial(
+                _sample_summary,
+                out_base=self.results_config.phy,
+                node_names=self.node_names, cols=self._cols, index=self.index,
+                cutoff=self.cutoff, cpm=self.cpm, sep=self.SEP
+            )
+            mp_pool.starmap_async(f, raw_sample_series).get()
+
+
+        ###
+        ### Save summarised results to disk.
+        (_, sl_series_list, ml_series_list) = zip(*raw_sample_series)
+        sldf = pd.concat(sl_series_list, axis=1)
+        mldf = pd.concat(ml_series_list, axis=1)
+
+        # Remove null rows.
+        if not self.keep_zeros:
+            (sldf, mldf) = remove_zeros(sldf, mldf)
+
+        # Drop unclassified from split DataFrame.
+        try:
+            mldf.drop(index="unclassified", inplace=True)
+        except KeyError:    # If no unclassified, this will have already been removed.
+            pass
+
+        # Save counts to disk.
+        self.save_df(sldf, self.results_config.phy_classified)
+        self.save_df(mldf, self.results_config.phy_split)
+
+    def save_df(self, df: pd.DataFrame, out_file: str):
+        self.fix_df_index(df)
+        df.to_csv(out_file, sep=self.SEP)
+
+    def fix_df_index(self, df: pd.DataFrame):
+        fixer = lambda name: 'p' + name if (name != "unclassified") and (self.index[name].type == "Branch") else name
+        df.index = [fixer(n) for n in df.index]
 
     @staticmethod
-    def get_data(file_url):
-        with open(file_url, "r") as f:
-            data = f.read().strip()
+    def concatenate_files(dest, *files):
+        with open(dest, 'w') as out_f:
+            for file in files:
+                # Write data to concatenated file.
+                with open(file, 'r') as in_f:
+                    out_f.write(in_f.read().strip() + "\n")
 
-        # Delete temporary file.
-        os.remove(file_url)
-
-        return data
+                # Remove temporary file.
+                os.remove(file)
 
     @staticmethod
     def best_match(file_name, file_list):
@@ -393,6 +408,8 @@ class Distribution:
         return match_ind
 
     def prepare_queue(self, paired_end=False):
+        file_queue = []
+
         for path in self.read_paths:
             # Create a list of all the files containing reads to be processed.
             file_list = [
@@ -401,8 +418,6 @@ class Distribution:
                 if re.match(r"(\S+)\.(?:fa|fna|ffn|fq|fasta|fastq)(?:\.tar)*(?:\.gz)*$", file_name)
             ]
 
-            file_queue = []
-
             if paired_end:  # This implies path is a folder, not a file.
                 while file_list:
                     next_file = file_list.pop()
@@ -411,14 +426,10 @@ class Distribution:
                     match_ind = self.best_match(next_file, file_list)
                     paired_file = file_list.pop(match_ind)
 
-                    file_queue.append((os.path.join(path, next_file), os.path.join(path, paired_file)))
-
+                    file_queue.append((next_file, paired_file))
             else:
                 for file_name in file_list:
-                    if os.path.isfile(file_name):
-                        file_queue.append((file_name, ))
-                    else:
-                        file_queue.append((os.path.join(path, file_name),))
+                    file_queue.append((file_name, ))
 
         return file_queue
 
@@ -440,9 +451,18 @@ class Distribution:
 
 
 class ClassificationResults:
+    _c_perc = "Cumulative Classified Percentage"
+    _c_cumul = "Cumulative Classified Count"
+    _c_count = "Raw Classified Count"
+    _s_perc = "Cumulative Split Percentage"
+    _s_cumul = "Cumulative Split Count"
+    _s_count = "Raw Split Count"
+    _rank = "Rank"
+    _scientific_name = "Scientific Name"
+
     def __init__(self, index, phylogeny_index, results_config, groups=None, keep_zeros=False, cutoff=0.0, cpm=0.0,
                  use_node_names=True, phyla=False, name_taxa=None, colour_list=None, circle_scale=1.0,
-                 log_scores=False):
+                 log_scores=False, sep=","):
         self.index = index
         self.phylogeny_index: Index = phylogeny_index
         self.results_config: ResultsPathConfig = results_config
@@ -463,14 +483,16 @@ class ClassificationResults:
         if self.phyla and self.name_taxa is None:
             raise Exception("Inclusion of phyla requires mapping of names to taxonomic lineages!")
 
-        self.tax_id_hierarchy = {"1": set()}  # Map from tax_id -> immediate children.
+        self.taxid_immediate_children = {"1": set()}  # Map from tax_id -> immediate children.
         self.tax_id_pool = ["1"]  # Children must appear later than parent this list.
 
-    def to_taxonomy(self, name_to_lineage, taxon_to_rank):
-        col_names = ["c_perc", "c_cumul", "c_count", "s_perc", "s_cumul", "s_count", "rank", "scientific name"]
+        self.SEP = sep
 
-        class_counts = pd.read_csv(self.results_config.phy_classified, sep="\t", index_col=0, header=0)
-        split_counts = pd.read_csv(self.results_config.phy_split, sep="\t", index_col=0, header=0)
+    def to_taxonomy(self, name_to_lineage, taxon_to_rank):
+        col_names = [self._c_perc, self._c_cumul, self._c_count, self._s_perc, self._s_cumul, self._s_count, self._rank, self._scientific_name]
+
+        class_counts = pd.read_csv(self.results_config.phy_classified, sep=self.SEP, index_col=0, header=0)
+        split_counts = pd.read_csv(self.results_config.phy_split, sep=self.SEP, index_col=0, header=0)
 
         # Get rid of phylogenetically printed node names.
         def fix_index(df: pd.DataFrame):
@@ -504,55 +526,57 @@ class ClassificationResults:
             ranks.append(rank)
             lineages.append(lineage)
 
+        # Make taxonomic output folder structure.
+        create_tax_results(self.results_config)
+
         for sample_name in samples:
             #
             # Map summary output to taxonomy.
             #
 
             df = pd.DataFrame(0, index=["unclassified"] + tax_ids, columns=col_names)
-            total_counts = sum(class_counts[sample_name]) + sum(split_counts[sample_name])
+            total_counts = class_counts[sample_name].sum() + split_counts[sample_name].sum()
 
-            df.loc[tax_ids[0]:, "rank"] = ranks
-            df.loc[tax_ids[0]:, "scientific name"] = lineages
+            df.loc[tax_ids[0]:, self._rank] = ranks
+            df.loc[tax_ids[0]:, self._scientific_name] = lineages
 
             # Manually insert unclassified counts.
-            df.loc["unclassified", "c_count"] = class_counts.loc["unclassified", sample_name]
-            df.loc["unclassified", "c_cumul"] = df.loc["unclassified", "c_count"]
-            df.loc["unclassified", "c_perc"] = round(100 * df.loc["unclassified", "c_cumul"] / total_counts, 3)
+            df.loc["unclassified", self._c_count] = class_counts.loc["unclassified", sample_name]
+            df.loc["unclassified", self._c_cumul] = df.loc["unclassified", self._c_count]
+            df.loc["unclassified", self._c_perc] = round(100 * df.loc["unclassified", self._c_cumul] / total_counts, 3)
 
-            for col_prefix, counts_df in (("c_", class_counts), ("s_", split_counts)):
+            for col, counts_df in ((self._c_count, class_counts), (self._s_count, split_counts)):
                 for node in counts_df.index[1:]:
-                    df.loc[phylogeny_to_taxonomy[node], col_prefix + "count"] += counts_df.loc[node, sample_name]
+                    df.loc[phylogeny_to_taxonomy[node], col] += counts_df.loc[node, sample_name]
 
-            # Get cumulative counts.
+            # Make counts cumulative.
+            # We may not be able to use groupby, but we can iterate through the taxid
+            # pool once (in hierarchical order).
             for tax_id in self.tax_id_pool[::-1]:
-                for pre in ("c_", "s_"):
-                    df.loc[tax_id, pre + "cumul"] = df.loc[tax_id, pre + "count"] + sum([
-                        df.loc[child, pre + "cumul"]
-                        for child in self.tax_id_hierarchy[tax_id]
-                    ])
-                    df.loc[tax_id, pre + "perc"] = 100 * df.loc[tax_id, pre + "cumul"] / total_counts
+                for count_col, cumul_col in (
+                    (self._c_count, self._c_cumul),
+                    (self._s_count, self._s_cumul)
+                ):
+                    df.loc[tax_id, cumul_col] = df.loc[tax_id, count_col] + df.loc[list(self.taxid_immediate_children[tax_id]), cumul_col].sum(axis=0)
 
             # Format percentage signs.
-            df.loc[:, "c_perc"] = round(df["c_perc"], 3).map(str) + "%"
-            df.loc[:, "s_perc"] = round(df["s_perc"], 3).map(str) + "%"
+            df.loc[:, self._c_perc] = round(100 * df.loc[:, self._c_cumul] / total_counts, 3).map(str) + "%"
+            df.loc[:, self._s_perc] = round(100 * df.loc[:, self._s_cumul] / total_counts, 3).map(str) + "%"
 
             # Employ cutoff.
             cutoff = max(self.cutoff, (total_counts / 1e6) * self.cpm)
-            df = df[(df['c_cumul'] > cutoff) | (df['s_cumul'] > cutoff) | (df.index == 'unclassified')]
-
-            df.to_csv(os.path.join(self.results_config.tax, sample_name + ".csv"), sep="\t", header=True)
+            df = df[(df[self._c_cumul] > cutoff) | (df[self._s_cumul] > cutoff) | (df.index == 'unclassified')]
+            df.to_csv(os.path.join(self.results_config.tax, sample_name + ".csv"), sep=self.SEP, header=True)
 
             #
             # Map raw read output to taxonomy.
             #
-
             raw_counts_file = os.path.join(self.results_config.phy_raw, sample_name + ".csv")
-            raw_read_data = []
+            output_file = os.path.join(self.results_config.tax_raw, sample_name + '.csv')
 
-            with open(raw_counts_file, 'r') as f:
-                for line in f:
-                    read_data = line.strip().split('\t')[:-1]
+            with open(raw_counts_file, 'r') as in_f, open(output_file, 'w') as out_f:
+                for line in in_f:
+                    read_data = line.strip().split("\t")[:-1]
 
                     if read_data[0] == 'U':
                         read_data[2] = 'NA'
@@ -561,14 +585,16 @@ class ClassificationResults:
                         phylogeny_node = self.phylogeny_index.pool[phy_id].name
                         read_data[2] = phylogeny_to_taxonomy[phylogeny_node]
 
-                    raw_read_data.append('\t'.join(read_data))
-
-            with open(os.path.join(self.results_config.tax_raw, sample_name + '.csv'), 'w') as f:
-                f.write('\n'.join(raw_read_data))
+                    out_f.write("\t".join(read_data) + "\n")
 
     def map_phylogeny_node(self, node_name, name_to_lineage, taxon_to_rank):
-        # Map from (node_name) --> (taxid, rank, taxonomic lineage)
-        #
+        """
+        Create a map from (node_name) --> (taxid, rank, taxonomic lineage).
+        
+        :param node_name: Name of clade.
+        :param name_to_lineage: Dict[Accession ID, Tuple[Taxonomic Lineage]]
+        :param taxon_to_rank: Dict[Taxonomic Name, Tuple[TaxID, Rank]]
+        """
         leaves = self.get_leaves(node_name)
         lineages = [name_to_lineage[leaf] for leaf in leaves]
         common_lineage = self.tuple_intersect(*lineages)
@@ -578,34 +604,38 @@ class ClassificationResults:
         else:
             phy_tax_id, phy_rank = taxon_to_rank[common_lineage[-1]]
 
-        tax_id_map = {}
+        taxid_rank_and_lineage = {}
         pool_index = 1
         for i in range(len(common_lineage)):
+            # Store taxonomic data.
             parent_tax_id, parent_rank = taxon_to_rank[common_lineage[i]]
-            tax_id_map[parent_tax_id] = [parent_rank, " ".join(common_lineage[:i + 1])]
-
+            taxid_rank_and_lineage[parent_tax_id] = [parent_rank, " ".join(common_lineage[:i + 1])]
+            
+            # Add top node as child of root.
             if i == 0:
-                # Connect to root.
-                self.tax_id_hierarchy["1"].add(parent_tax_id)
+                self.taxid_immediate_children["1"].add(parent_tax_id)
 
-            # Set tax_id in pool.
+            # Set tax_id in pool - a list of taxids such that child taxids are necessarily further
+            # in the last than any ancestors.
             try:
                 pool_index = self.tax_id_pool.index(parent_tax_id) + 1
             except ValueError:
                 self.tax_id_pool.insert(pool_index, parent_tax_id)
                 pool_index += 1
 
-            # Set taxa in hierarchy.
+            # Store immediate child of taxid.
             if i + 1 < len(common_lineage):
                 child_tax_id, child_rank = taxon_to_rank[common_lineage[i + 1]]
-                if parent_tax_id in self.tax_id_hierarchy:
-                    self.tax_id_hierarchy[parent_tax_id].add(child_tax_id)
-                else:
-                    self.tax_id_hierarchy[parent_tax_id] = {child_tax_id}
-            elif parent_tax_id not in self.tax_id_hierarchy:
-                self.tax_id_hierarchy[parent_tax_id] = set()
 
-        return phy_tax_id, tax_id_map
+                try:
+                    self.taxid_immediate_children[parent_tax_id].add(child_tax_id)
+                except KeyError:
+                    self.taxid_immediate_children[parent_tax_id] = {child_tax_id}
+
+            elif parent_tax_id not in self.taxid_immediate_children:
+                self.taxid_immediate_children[parent_tax_id] = set()
+
+        return phy_tax_id, taxid_rank_and_lineage
 
     def get_leaves(self, node_name):
         return self.phylogeny_index.get_child_leaves(node_name)
@@ -642,7 +672,8 @@ class ClassificationResults:
             keep_zeros=self.keep_zeros,
             use_node_names=self.use_node_names,
             log_scores=self.log_scores,
-            itol_mode=itol_mode
+            itol_mode=itol_mode,
+            sep=self.SEP
         )
 
         # Draw unclassified tree.
@@ -658,5 +689,6 @@ class ClassificationResults:
             keep_zeros=self.keep_zeros,
             use_node_names=self.use_node_names,
             log_scores=self.log_scores,
-            itol_mode=itol_mode
+            itol_mode=itol_mode,
+            sep=self.SEP
         )
